
#include "gdal_pam.h"
#include "fgdreader.h"

CPL_C_START
void	GDALRegister_FGD(void);
CPL_C_END

class FGDRasterBand;

class FGDDataset : public GDALPamDataset
{
	friend class FGDRasterBand;

public:
	FGDDataset();
	~FGDDataset();

	static GDALDataset* Open( GDALOpenInfo *poOpenInfo );
	static int Identify( GDALOpenInfo * poOpenInfo  );

	CPLErr GetGeoTransform( double *padfTransform );
	const char *GetProjectionRef();

private:
	GDALFGDReader* m_pReader;
};

class FGDRasterBand : public GDALPamRasterBand
{
    friend class FGDDataset;

public:

	~FGDRasterBand();
	FGDRasterBand( FGDDataset *poDS, int nBand );
    
    virtual CPLErr IReadBlock( int nBlockXOff, int nBlockYOff, void *pImage );
	virtual double GetNoDataValue( int *pbSuccess );
};


/************************************************************************/
/*                           FGDRasterBand()                            */
/************************************************************************/

FGDRasterBand::FGDRasterBand( FGDDataset *poDS, int nBand )
{
	this->poDS = poDS;
	this->nBand = nBand;

	eDataType = GDT_Float32;
	nBlockXSize = poDS->GetRasterXSize();
	nBlockYSize = 1;
}


/************************************************************************/
/*                           ~FGDRasterBand()                           */
/************************************************************************/

FGDRasterBand::~FGDRasterBand()
{
}


/************************************************************************/
/*                             IReadBlock()                             */
/************************************************************************/

CPLErr FGDRasterBand::IReadBlock( int nBlockXOff, int nBlockYOff, void *pImage )
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)

	float fPix;
	FGDDataset *poGDS = (FGDDataset *)poDS;

	poGDS->m_pReader->setLoi( nBlockYOff );

	for ( int i = 0; i < nBlockXSize; i++ )
	{
		if ( !poGDS->m_pReader->GetNextPix( &fPix ) )
		{
			((float *)pImage)[i] = FGD_NODATA;
		}
		else
		{
			((float *)pImage)[i] = fPix;
		}
	}

	return CE_None;

#else
	CPLError(CE_Failure, CPLE_NotSupported,
		"GDAL/FGD driver has not been built with read support. Expat and ICONV library required");

	return CE_Failure;
#endif
}


/************************************************************************/
/*                           GetNoDataValue()                           */
/************************************************************************/
double FGDRasterBand::GetNoDataValue( int *pbSuccess )
{
	if ( pbSuccess != NULL )
		*pbSuccess = TRUE;

	return FGD_NODATA;
}



/************************************************************************/
/*                             FGDDataset()                             */
/************************************************************************/

FGDDataset::FGDDataset()
{
	m_pReader = NULL;
}



/************************************************************************/
/*                             ~FGDDataset()                            */
/************************************************************************/

FGDDataset::~FGDDataset()
{
	if ( m_pReader )
		delete m_pReader;
}



/************************************************************************/
/*                          GetGeoTransform()                           */
/************************************************************************/

CPLErr FGDDataset::GetGeoTransform( double *padfTransform )
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	if ( !m_pReader )
	{
		CPLError( CE_Failure, CPLE_OpenFailed,
			"File open failed." );
		return CE_Failure;
	}

	if ( !m_pReader->isValid() )
	{
		CPLError( CE_Failure, CPLE_AppDefined,
			"Invalid FGD File." );
		return CE_Failure;
	}

	m_pReader->GetGeoTransform( padfTransform );

	return CE_None;

#else
	CPLError(CE_Failure, CPLE_NotSupported,
		"GDAL/FGD driver has not been built with read support. Expat and ICONV library required");

	return CE_Failure;
#endif
}



/************************************************************************/
/*                          GetProjectionRef()                          */
/************************************************************************/

const char *FGDDataset::GetProjectionRef()
{
		return "GEOGCS[\"JGD2000\", "
        "   DATUM[\"Japanese_Geodetic_Datum_2000\","
        "       SPHEROID[\"GRS 1980\",6378137,298.257222101,"
        "           AUTHORITY[\"EPSG\",\"7019\"]],"
        "       TOWGS84[0,0,0,0,0,0,0],"
        "       AUTHORITY[\"EPSG\",\"6612\"]],"
        "   PRIMEM[\"Greenwich\",0,"
        "       AUTHORITY[\"EPSG\",\"8901\"]],"
        "   UNIT[\"degree\",0.0174532925199433,"
        "       AUTHORITY[\"EPSG\",\"9122\"]],"
        "   AUTHORITY[\"EPSG\",\"4612\"]]";
}



/************************************************************************/
/*                              Identify()                              */
/************************************************************************/

int FGDDataset::Identify( GDALOpenInfo * poOpenInfo )
{
	const char *pszExtension = CPLGetExtension( poOpenInfo->pszFilename );

	if ( !EQUAL( pszExtension, "xml" ) )
	{
		return false;
	}

    if( poOpenInfo->nHeaderBytes > 20
         && strstr((const char *)poOpenInfo->pabyHeader,"<GI") != NULL )
        return true;

    if( strstr(poOpenInfo->pszFilename,"<GI") != NULL )
        return true;

    return false;
}



/************************************************************************/
/*                                Open()                                */
/************************************************************************/
GDALDataset *FGDDataset::Open( GDALOpenInfo * poOpenInfo )
{
	FGDDataset *poDS = NULL;

	if ( !Identify( poOpenInfo ) )
		return NULL;

	if ( poOpenInfo->eAccess == GA_Update )
	{
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "The FGD driver does not support update access to existing"
                  " datasets.\n" );
        return NULL;
	}

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	poDS = new FGDDataset();

	poDS->m_pReader = new GDALFGDReader( poOpenInfo->pszFilename );
	if ( !poDS->m_pReader->isOpen() )
	{
		return NULL;
	}

	if ( !poDS->m_pReader->isValid() )
	{
		return NULL;
	}

	poDS->nRasterXSize = poDS->m_pReader->GetXSize();
	poDS->nRasterYSize = poDS->m_pReader->GetYSize();

	poDS->SetBand( 1, new FGDRasterBand( poDS, 1 ) );

	poDS->SetDescription( poOpenInfo->pszFilename );
	poDS->TryLoadXML();

	poDS->oOvManager.Initialize( poDS, poOpenInfo->pszFilename );
#endif

	return poDS;
}


/************************************************************************/
/*                          GDALRegister_FGD()                          */
/************************************************************************/

void GDALRegister_FGD()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	GDALDriver *poDriver;

	if ( GDALGetDriverByName( "FGD" ) == NULL )
	{
        poDriver = new GDALDriver();

		poDriver->SetDescription( "FGD" );
		poDriver->SetMetadataItem( GDAL_DMD_LONGNAME,
			"Japanese Fundamental Geographic Data (.xml)" );
		poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC,
			"frmt_various.html#FGD" );
		poDriver->SetMetadataItem( GDAL_DMD_EXTENSION, "xml" );
		poDriver->SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" );

		poDriver->pfnOpen = FGDDataset::Open;
		poDriver->pfnIdentify = FGDDataset::Identify;

		GetGDALDriverManager()->RegisterDriver( poDriver );
	}
#endif
}

