
#include "fgdreader.h"
#include <cpl_conv.h>
#include <cpl_string.h>
#include <cpl_port.h>


#if defined(__GNUC__)
	#define STRCMPI strcasecmp
#elif defined(WIN32)
	#define STRCMPI stricmp
#else
	#define STRCMPI strcmp
#endif


#ifndef ICONV_CPP_CONST
#define ICONV_CPP_CONST ICONV_CONST
#endif

#ifdef HAVE_EXPAT


static void XMLCALL startElement( void *pUserData, const char *pszName, const char **ppszAttr )
{
	FGDReader* poDS = (FGDReader*) pUserData;
	poDS->startElement( pszName, ppszAttr );
}

static void XMLCALL endElement( void *pUserData, const char *pszName )
{
	FGDReader* poDS = (FGDReader*) pUserData;
	poDS->endElement( pszName );
}

static int convertSJIStoUTF8( void *data, const char *c )
{
	FGDReader *poDS = (FGDReader *)data;
	return poDS->convertSJIStoUTF8( c );
}

static void XMLCALL dataHandler( void *pUserData, const char *pszData, int nLen )
{
	FGDReader* poDS = (FGDReader*) pUserData;
	poDS->dataHandler( pszData, nLen );
}

static int XML_ShiftJISEncodingHandler( void *encodingHandlerData, const XML_Char *name, XML_Encoding *info )
{
	int i;

	if ( STRCMPI( name, "shift-jis" ) && STRCMPI( name, "shift_jis" ) )
	{
		return 0;
	}

	// ASCII Map
	info->map['\t'] = '\t';
	info->map['\r'] = '\r';
	info->map['\n'] = '\n';

	for ( i = ' '; i <= '~'; i++ )
	{
		info->map[i] = i;
	}

	for ( i = 0x81; i <= 0x9f; i++ )
	{
		info->map[i] = -2;
	}
	for ( i = 0xa1; i <= 0xdf; i++ )
	{
		info->map[i] = i + 0xfec0;
	}
	for ( i = 0xe0; i <= 0xfc; i++ )
	{
		info->map[i] = -2;
	}

	info->data = encodingHandlerData;
	info->convert = convertSJIStoUTF8;
	info->release = NULL;

	return 1;
}

#endif

FGDReader::FGDReader( const char *pszFilename )
{
	m_bOpen = false;
	m_poSRS = new OGRSpatialReference("GEOGCS[\"JGD2000\", "
        "   DATUM[\"Japanese_Geodetic_Datum_2000\","
        "       SPHEROID[\"GRS 1980\",6378137,298.257222101,"
        "           AUTHORITY[\"EPSG\",\"7019\"]],"
        "       TOWGS84[0,0,0,0,0,0,0],"
        "       AUTHORITY[\"EPSG\",\"6612\"]],"
        "   PRIMEM[\"Greenwich\",0,"
        "       AUTHORITY[\"EPSG\",\"8901\"]],"
        "   UNIT[\"degree\",0.0174532925199433,"
        "       AUTHORITY[\"EPSG\",\"9122\"]],"
        "   AUTHORITY[\"EPSG\",\"4612\"]]");

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	m_nNextFID = 0;
	m_validity = FGD_VALIDITY_UNKNOWN;

	m_nFlags = 0x0000;

	m_poPoint = NULL;
	m_poLineString = NULL;
	m_poRing = NULL;
	m_poPolygon = NULL;

	fp = VSIFOpenL( pszFilename, "r" );
	if ( fp == NULL )
		return;

	m_nDataHandlerCounter = 0;
	m_poFeatureDefn = new OGRFeatureDefn( CPLGetBasename( pszFilename ) );

	m_Parser = XML_ParserCreate( NULL );
	XML_SetUserData( m_Parser, this );
	XML_SetElementHandler( m_Parser, ::startElement, ::endElement );
	XML_SetCharacterDataHandler( m_Parser, ::dataHandler );
	XML_SetUnknownEncodingHandler( m_Parser, XML_ShiftJISEncodingHandler, this );

	m_hIconv = iconv_open( "UTF-16BE", "CP932" );

	while ( m_validity == FGD_VALIDITY_UNKNOWN )
	{
		if ( !parseNext() )
		{
			m_validity = FGD_VALIDITY_INVALID;
			return;
		}
	}

	if ( m_validity == FGD_VALIDITY_INVALID )
	{
		return;
	}

	while ( m_poFeatureDefn->GetGeomType() == wkbUnknown )
	{
		if ( !parseNext() )
		{
			m_validity = FGD_VALIDITY_INVALID;
			return;
		}
	}

	ResetReading();
#endif
}



FGDReader::~FGDReader()
{
#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
	XML_ParserFree( m_Parser );

	iconv_close( m_hIconv );
	if ( fp )
		VSIFCloseL( fp );

#endif

	m_poSRS->Release();
	m_bOpen = false;
}

#if defined(HAVE_EXPAT) && defined(HAVE_ICONV)


int FGDReader::convertSJIStoUTF8( const char *s )
{
	char *pszFrom, *pszTo;
	char szFrom[2], szTo[BUFSIZ];
	size_t nInByte, nOutByte;
	int nByte;
	unsigned short nChar = 0;
	ICONV_CPP_CONST char *pszSrcBuf = (ICONV_CONST char *)s;

	if ( !m_hIconv )
		return 0;

	memcpy( szFrom, s, 2 );
	pszFrom = szFrom;
	pszTo = szTo;
	nInByte = 2;
	nOutByte = nInByte * 2;
	iconv( m_hIconv, &pszSrcBuf, &nInByte, &pszTo, &nOutByte );
	*pszTo = '\0';

	nByte = pszTo - szTo;
	if ( nByte > 4 ) return 0;
	pszTo = szTo;
	if ( nByte == 4 )
		pszTo += 2;
	nChar = (*pszTo << 8) | (*(pszTo+1) & 0x00ff);

	return (int)nChar;
}


bool FGDReader::setLayerDefn( const char *pszName )
{
	OGRFieldDefn oFidDefn( "fid", OFTString );
	m_poFeatureDefn->AddFieldDefn( &oFidDefn );
	OGRFieldDefn oLfSpanFr( "lfSpanFr", OFTDate );
	m_poFeatureDefn->AddFieldDefn( &oLfSpanFr );
	OGRFieldDefn oLfSpanTo( "lfSpanTo", OFTDate );
	m_poFeatureDefn->AddFieldDefn( &oLfSpanTo );
	OGRFieldDefn oDevDate( "devDate", OFTDate );
	m_poFeatureDefn->AddFieldDefn( &oDevDate );
	OGRFieldDefn oOrgGILvl( "orgGILvl", OFTInteger );
	m_poFeatureDefn->AddFieldDefn( &oOrgGILvl );
	OGRFieldDefn oOrgMDId( "orgMDId", OFTString );
	m_poFeatureDefn->AddFieldDefn( &oOrgMDId );
	OGRFieldDefn oVis( "vis", OFTString );
	m_poFeatureDefn->AddFieldDefn( &oVis );

	if ( !STRCMPI( pszName, "FGDFeature" ) )
	{
		m_poFeatureDefn->SetGeomType( wkbUnknown );
		return false;
	}
	else if ( !STRCMPI( pszName, "GCP" ) )
	{
		OGRFieldDefn oAdvNo( "advNo", OFTInteger );
		m_poFeatureDefn->AddFieldDefn( &oAdvNo );
		OGRFieldDefn oOrgName( "orgName", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oOrgName );
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oGcpClass( "gcpClass", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oGcpClass );
		OGRFieldDefn oGcpCode( "gcpCode", OFTInteger );
		m_poFeatureDefn->AddFieldDefn( &oGcpCode );
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );
		OGRFieldDefn oB( "B", OFTReal );
		m_poFeatureDefn->AddFieldDefn( &oB );
		OGRFieldDefn oL( "L", OFTReal );
		m_poFeatureDefn->AddFieldDefn( &oL );
		OGRFieldDefn oAlti( "alti", OFTReal );
		m_poFeatureDefn->AddFieldDefn( &oAlti );
		OGRFieldDefn oAltiAcc( "altiAcc", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oAltiAcc );

		m_poFeatureDefn->SetGeomType( wkbPoint );

		return true;
	}
	else if ( !STRCMPI( pszName, "ElevPt" ) ||
			    !STRCMPI( pszName, "DEMPt" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oAlti( "alti", OFTReal );
		m_poFeatureDefn->AddFieldDefn( &oAlti );

		m_poFeatureDefn->SetGeomType( wkbPoint );

		return true;
	}
	else if ( !STRCMPI( pszName, "AdmPt" ) ||
				!STRCMPI( pszName, "CommPt" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );
		OGRFieldDefn oAdmCode( "admCode", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oAdmCode );

		m_poFeatureDefn->SetGeomType( wkbPoint );

		return true;
	}
	else if ( !STRCMPI( pszName, "SBAPt" ) )
	{
		OGRFieldDefn oSbNo( "sbNo", OFTInteger );
		m_poFeatureDefn->AddFieldDefn( &oSbNo );

		m_poFeatureDefn->SetGeomType( wkbPoint );

		return true;
	}
	else if ( !STRCMPI( pszName, "Cntr" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oAlti( "alti", OFTReal );
		m_poFeatureDefn->AddFieldDefn( &oAlti );

		m_poFeatureDefn->SetGeomType( wkbLineString );

		return true;
	}
	else if ( !STRCMPI( pszName, "AdmBdry" ) ||
				!STRCMPI( pszName, "CommBdry" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );

		m_poFeatureDefn->SetGeomType( wkbLineString );

		return true;
	}
	else if ( !STRCMPI( pszName, "SBBdry" ) ||
				!STRCMPI( pszName, "RdASL" ) )
	{
		m_poFeatureDefn->SetGeomType( wkbLineString );

		return true;
	}
	else if ( !STRCMPI( pszName, "Cstline" ) ||
				!STRCMPI( pszName, "WL" ) ||
				!STRCMPI( pszName, "WStrL" ) ||
				!STRCMPI( pszName, "WStrA" ) ||
				!STRCMPI( pszName, "BldL" ) ||
				!STRCMPI( pszName, "RdEdg" ) ||
				!STRCMPI( pszName, "RdCompt" ) ||
				!STRCMPI( pszName, "RailCL" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );

		m_poFeatureDefn->SetGeomType( wkbLineString );

		return true;
	}
	else if ( !STRCMPI( pszName, "RvrMgtBdry" ) ||
				!STRCMPI( pszName, "LeveeEdge" ) ||
				!STRCMPI( pszName, "RdMgtBdry" ) )
	{
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );

		m_poFeatureDefn->SetGeomType( wkbLineString );

		return true;
	}
	else if ( !STRCMPI( pszName, "AdmArea" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );
		OGRFieldDefn oAdmCode( "admCode", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oAdmCode );

		m_poFeatureDefn->SetGeomType( wkbPolygon );

		return true;
	}
	else if ( !STRCMPI( pszName, "SBArea" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oSbNo( "sbNo", OFTInteger );
		m_poFeatureDefn->AddFieldDefn( &oSbNo );

		m_poFeatureDefn->SetGeomType( wkbPolygon );

		return true;
	}
	else if ( !STRCMPI( pszName, "WA" ) ||
				!STRCMPI( pszName, "BldA" ) ||
				!STRCMPI( pszName, "RdSgmtA" ) )
	{
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );

		m_poFeatureDefn->SetGeomType( wkbPolygon );

		return true;
	}
	else if ( !STRCMPI( pszName, "RdArea" ) )
	{
		OGRFieldDefn oName( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oName );

		m_poFeatureDefn->SetGeomType( wkbPolygon );

		return true;
	}
	else if ( !STRCMPI( pszName, "DEM" ) )
	{
/*
		OGRFieldDefn oType( "type", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oType );
		OGRFieldDefn oMesh( "name", OFTString );
		m_poFeatureDefn->AddFieldDefn( &oMesh );

		m_poFeatureDefn->SetGeomType( OGRwkbGeometryType::wkbPoint );
*/
		CPLError( CE_Failure, CPLE_NotSupported, "FGD-DEM format is not supported." );

		return false;
	}
	else
	{
		CPLError( CE_Failure, CPLE_AppDefined, "Unknown format." );

		return false;
	}
}

void FGDReader::startElement(const char *pszName, const char **ppszAttr)
{
	m_szTempChar[0] = '\0';
	m_nTempChar = 0;

	if ( !STRCMPI( pszName, "GI" ) )
	{
		int i;
		m_nFlags |= TAG_GI;
		for ( i = 0; ppszAttr[i] != NULL; i += 2 )
		{
			if ( !strcmp( ppszAttr[i], "version" ) )
			{
				m_pszVersion = CPLStrdup( ppszAttr[i+1] );
				return;
			}
		}
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags |= TAG_DATASET;
		if ( !( m_nFlags & TAG_GI ) )
		{
			m_validity = FGD_VALIDITY_INVALID;
		}
		return;
	}
	else if ( !STRCMPI( pszName, "FGDFeature" ) ||
		 !STRCMPI( pszName, "GCP" ) ||
		 !STRCMPI( pszName, "ElevPt" ) ||
		 !STRCMPI( pszName, "DEMPt" ) ||
		 !STRCMPI( pszName, "AdmPt" ) ||
		 !STRCMPI( pszName, "CommPt" ) ||
		 !STRCMPI( pszName, "SBAPt" ) ||
		 !STRCMPI( pszName, "Cntr" ) ||
		 !STRCMPI( pszName, "AdmBdry" ) ||
		 !STRCMPI( pszName, "CommBdry" ) ||
		 !STRCMPI( pszName, "SBBdry" ) ||
		 !STRCMPI( pszName, "RdASL" ) ||
		 !STRCMPI( pszName, "Cstline" ) ||
		 !STRCMPI( pszName, "WL" ) ||
		 !STRCMPI( pszName, "WStrL" ) ||
		 !STRCMPI( pszName, "WStrA" ) ||
		 !STRCMPI( pszName, "BldL" ) ||
		 !STRCMPI( pszName, "RdEdg" ) ||
		 !STRCMPI( pszName, "RdCompt" ) ||
		 !STRCMPI( pszName, "RdSgmtA" ) ||
		 !STRCMPI( pszName, "RailCL" ) ||
		 !STRCMPI( pszName, "RvrMgtBdry" ) ||
		 !STRCMPI( pszName, "LeveeEdge" ) ||
		 !STRCMPI( pszName, "RdMgtBdry" ) ||
		 !STRCMPI( pszName, "AdmArea" ) ||
		 !STRCMPI( pszName, "SBArea" ) ||
		 !STRCMPI( pszName, "WA" ) ||
		 !STRCMPI( pszName, "BldA" ) ||
		 !STRCMPI( pszName, "RdArea" ) ||
		 !STRCMPI( pszName, "DEM" ) )
	{
		m_nFlags |= TAG_ELEMENT;
		if ( m_validity == FGD_VALIDITY_UNKNOWN )
		{
			if ( !(m_nFlags & TAG_GI) || !(m_nFlags & TAG_DATASET) )
			{
				m_validity = FGD_VALIDITY_INVALID;
				return;
			}
			if ( m_poFeatureDefn->GetGeomType() == wkbUnknown )
			{
				if ( setLayerDefn( pszName ) )
				{
					m_validity = FGD_VALIDITY_VALID;
				}
				else
				{
					m_validity = FGD_VALIDITY_INVALID;
				}
			}
		}

		if ( m_validity == FGD_VALIDITY_VALID )
			m_poCurFeature = new OGRFeature( m_poFeatureDefn );

		return;
	}
	else if ( !STRCMPI( pszName, "fid" ) ||
			  !STRCMPI( pszName, "orgGILvl" ) ||
			  !STRCMPI( pszName, "orgMDId" ) ||
			  !STRCMPI( pszName, "vis" ) ||
			  !STRCMPI( pszName, "advNo" ) ||
			  !STRCMPI( pszName, "orgName" ) ||
			  !STRCMPI( pszName, "type" ) ||
			  !STRCMPI( pszName, "gcpClass" ) ||
			  !STRCMPI( pszName, "gcpCode" ) ||
			  !STRCMPI( pszName, "name" ) ||
			  !STRCMPI( pszName, "B" ) ||
			  !STRCMPI( pszName, "L" ) ||
			  !STRCMPI( pszName, "alti" ) ||
			  !STRCMPI( pszName, "altiAcc" ) ||
			  !STRCMPI( pszName, "admCode" ) ||
			  !STRCMPI( pszName, "sbNo" ) )
	{
		m_nFlags |= TAG_ATTRIBUTE;
		return;
	}
	else if ( !STRCMPI( pszName, "lfSpanFr" ) ||
			  !STRCMPI( pszName, "lfSpanTo" ) ||
			  !STRCMPI( pszName, "devDate" ) )
	{
		m_nFlags |= TAG_DATE_ATTRIBUTE;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:date8601" ) )
	{
		m_nFlags |= TAG_DATE;
		return;
	}
	else if ( !STRCMPI( pszName, "pos" ) )
	{
		m_nFlags |= TAG_POS;
		m_poPoint = new OGRPoint();
		m_poPoint->setCoordinateDimension( 2 );
		return;
	}
	else if ( !STRCMPI( pszName, "loc" ) )
	{
		m_poLineString = new OGRLineString();
		m_poLineString->setCoordinateDimension( 2 );
		m_nFlags |= TAG_LOC;
		return;
	}
	else if ( !STRCMPI( pszName, "area" ) )
	{
		m_poPolygon = new OGRPolygon();
		m_poPolygon->setCoordinateDimension( 2 );
		m_nFlags |= TAG_AREA;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:GM_LineString" ) )
	{
		m_nFlags |= TAG_LINESTRING;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:GM_Polygon" ) )
	{
		m_nFlags |= TAG_POLYGON;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:exterior" ) )
	{
		m_poRing = new OGRLinearRing();
		m_poRing->setCoordinateDimension( 2 );
		m_nFlags |= TAG_EXTERIOR_RING;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:interior" ) )
	{
		m_poRing = new OGRLinearRing();
		m_poRing->setCoordinateDimension( 2 );
		m_nFlags |= TAG_INTERIOR_RING;
		return;
	}
	else if ( !STRCMPI( pszName, "jps:coordinate" ) )
	{
		m_nFlags |= TAG_COORDINATE;
		return;
	}
}


void FGDReader::endElement(const char *pszName)
{
	if ( !STRCMPI( pszName, "GI" ) )
	{
		m_nFlags &= ~TAG_GI;
		return;
	}
	else if ( !STRCMPI( pszName, "dataset" ) )
	{
		m_nFlags &= ~TAG_DATASET;
		return;
	}

	if ( m_validity == FGD_VALIDITY_VALID )
	{
		if ( !STRCMPI( pszName, "GCP" ) ||
			 !STRCMPI( pszName, "ElevPt" ) ||
			 !STRCMPI( pszName, "DEMPt" ) ||
			 !STRCMPI( pszName, "AdmPt" ) ||
			 !STRCMPI( pszName, "CommPt" ) ||
			 !STRCMPI( pszName, "SBAPt" ) )
		{
			m_nFlags &= ~TAG_ELEMENT;

			m_poPoint->assignSpatialReference( m_poSRS );
			m_poCurFeature->SetGeometryDirectly( m_poPoint );
			m_poCurFeature->SetFID( m_nNextFID++ );
			m_dqpoFeature.push_back( m_poCurFeature );

			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
			return;
		}
		else if ( !STRCMPI( pszName, "Cntr" ) ||
				  !STRCMPI( pszName, "AdmBdry" ) ||
				  !STRCMPI( pszName, "CommBdry" ) ||
				  !STRCMPI( pszName, "SBBdry" ) ||
				  !STRCMPI( pszName, "RdASL" ) ||
				  !STRCMPI( pszName, "Cstline" ) ||
				  !STRCMPI( pszName, "WL" ) ||
				  !STRCMPI( pszName, "WStrL" ) ||
				  !STRCMPI( pszName, "BldL" ) ||
				  !STRCMPI( pszName, "RdEdg" ) ||
				  !STRCMPI( pszName, "RdCompt" ) ||
				  !STRCMPI( pszName, "RailCL" ) ||
				  !STRCMPI( pszName, "RvrMgtBdry" ) ||
				  !STRCMPI( pszName, "LeveeEdge" ) ||
				  !STRCMPI( pszName, "RdMgtBdry" ) )
		{
			m_nFlags &= ~TAG_ELEMENT;

			m_poLineString->assignSpatialReference( m_poSRS );
			m_poCurFeature->SetGeometryDirectly( m_poLineString );
			m_poCurFeature->SetFID( m_nNextFID++ );
			m_dqpoFeature.push_back( m_poCurFeature );

			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
			return;
		}
		else if ( !STRCMPI( pszName, "AdmArea" ) ||
				  !STRCMPI( pszName, "SBArea" ) ||
				  !STRCMPI( pszName, "WA" ) ||
				  !STRCMPI( pszName, "WStrA" ) ||
				  !STRCMPI( pszName, "BldA" ) ||
				  !STRCMPI( pszName, "RdArea" ) ||
				  !STRCMPI( pszName, "RdSgmtA" ) )
		{
			m_nFlags &= ~TAG_ELEMENT;

			m_poPolygon->assignSpatialReference( m_poSRS );
			m_poCurFeature->SetGeometryDirectly( m_poPolygon );
			m_poCurFeature->SetFID( m_nNextFID++ );
			m_dqpoFeature.push_back( m_poCurFeature );

			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
			return;
		}
		else if ( !STRCMPI( pszName, "DEM" ) )
		{
			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
			m_nFlags &= ~TAG_ELEMENT;
		}
		else if ( !STRCMPI( pszName, "jps:date8601" ) )
		{
			m_nFlags &= ~TAG_DATE;
			return;
		}
		else if ( !STRCMPI( pszName, "jps:GM_Polygon" ) )
		{
			m_nFlags &= ~TAG_POLYGON;
			return;
		}
		else if ( !STRCMPI( pszName, "pos" ) )
		{
			m_nFlags &= ~TAG_POS;
			return;
		}
		else if ( !STRCMPI( pszName, "loc" ) )
		{
			m_nFlags &= ~TAG_LOC;
			return;
		}
		else if ( !STRCMPI( pszName, "area" ) )
		{
			m_nFlags &= ~TAG_AREA;
			return;
		}
		else if ( !STRCMPI( pszName, "jps:GM_LineString" ) )
		{
			m_nFlags &= ~TAG_LINESTRING;
			return;
		}
		else if ( !STRCMPI( pszName, "jps:GM_Polygon" ) )
		{
			m_nFlags &= ~TAG_POLYGON;
			return;
		}
		else if ( !STRCMPI( pszName, "jps:exterior" ) )
		{
			m_nFlags &= ~TAG_EXTERIOR_RING;
			m_poPolygon->addRingDirectly( m_poRing );
			for ( int i = 0; i < m_vpoInteriorRings.size(); i++ )
				m_poPolygon->addRingDirectly( m_vpoInteriorRings[i] );
			m_vpoInteriorRings.clear();
			return;
		}
		else if ( !STRCMPI( pszName, "jps:interior" ) )
		{
			m_nFlags &= ~TAG_INTERIOR_RING;
			if ( m_poPolygon->IsEmpty() )
				m_vpoInteriorRings.push_back( m_poRing );
			else
				m_poPolygon->addRingDirectly( m_poRing );

			return;
		}
		else if ( !STRCMPI( pszName, "jps:coordinate" ) )
		{
			float fLon, fLat;

			m_nFlags &= ~TAG_COORDINATE;
			sscanf( m_szTempChar, "%f %f", &fLat, &fLon );

			if ( m_nFlags & TAG_POS )
			{
				m_poPoint->setX( fLon );
				m_poPoint->setY( fLat );
			}
			else if ( m_nFlags & TAG_LOC )
			{
				m_poLineString->addPoint( fLon, fLat );
			}
			else
			{
				m_poRing->addPoint( fLon, fLat );
			}
			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
			return;
		}
		else if ( !STRCMPI( pszName, "fid" ) ||
				  !STRCMPI( pszName, "vis" ) ||
				  !STRCMPI( pszName, "orgName" ) ||
				  !STRCMPI( pszName, "type" ) ||
				  !STRCMPI( pszName, "gcpClass" ) ||
				  !STRCMPI( pszName, "name" ) ||
				  !STRCMPI( pszName, "altiAcc" ) ||
				  !STRCMPI( pszName, "admCode" ) ||
				  !STRCMPI( pszName, "orgMDId" ) ||
				  !STRCMPI( pszName, "lfSpanFr" ) ||
				  !STRCMPI( pszName, "lfSpanTo" ) ||
				  !STRCMPI( pszName, "devDate" ) ||
				  !STRCMPI( pszName, "orgGILvl" ) ||
				  !STRCMPI( pszName, "advNo" ) ||
				  !STRCMPI( pszName, "gcpCode" ) ||
				  !STRCMPI( pszName, "sbNo" ) ||
				  !STRCMPI( pszName, "B" ) ||
				  !STRCMPI( pszName, "L" ) ||
				  !STRCMPI( pszName, "alti" ) )
		{
			for ( int i = 0; i < m_poFeatureDefn->GetFieldCount(); i++ )
			{
				if ( !STRCMPI( m_poFeatureDefn->GetFieldDefn( i )->GetNameRef(), pszName ) )
				{
					if ( !STRCMPI( pszName, "fid" ) ||
						 !STRCMPI( pszName, "vis" ) ||
						 !STRCMPI( pszName, "orgName" ) ||
						 !STRCMPI( pszName, "type" ) ||
						 !STRCMPI( pszName, "gcpClass" ) ||
						 !STRCMPI( pszName, "name" ) ||
						 !STRCMPI( pszName, "altiAcc" ) ||
						 !STRCMPI( pszName, "admCode" ) ||
						 !STRCMPI( pszName, "orgMDId" ) )
					{
						if ( CPLStrnlen( m_szTempChar, MAX_LINES ) )
							m_poCurFeature->SetField( i, m_szTempChar );
						m_nFlags &= ~TAG_ATTRIBUTE;
						break;
					}
					else if ( !STRCMPI( pszName, "lfSpanFr" ) ||
							  !STRCMPI( pszName, "lfSpanTo" ) ||
							  !STRCMPI( pszName, "devDate" ) )
					{
						char **papszTokens = CSLTokenizeString2( m_szTempChar, "-", 0x00 );
						m_poCurFeature->SetField( i, atoi( papszTokens[0] ), atoi( papszTokens[1] ), atoi( papszTokens[2] ) );
						m_nFlags &= ~TAG_ATTRIBUTE;
						break;
					}
					else if ( !STRCMPI( pszName, "orgGILvl" ) ||
							  !STRCMPI( pszName, "advNo" ) ||
							  !STRCMPI( pszName, "gcpCode" ) ||
							  !STRCMPI( pszName, "sbNo" ) )
					{
						m_poCurFeature->SetField( i, atoi( m_szTempChar ) );
						m_nFlags &= ~TAG_ATTRIBUTE;
						break;
					}
					else if ( !STRCMPI( pszName, "B" ) ||
							  !STRCMPI( pszName, "L" ) ||
							  !STRCMPI( pszName, "alti" ) )
					{
						char *pszStop;
						double dData = strtod( m_szTempChar, &pszStop );
						m_poCurFeature->SetField( i, dData );
						m_nFlags &= ~TAG_ATTRIBUTE;
						break;
					}
				}
			}
			m_szTempChar[0] = '\0';
			m_nTempChar = 0;
		}
	}
}


void FGDReader::dataHandler(const char *data, int nLen)
{
	if ( m_validity == FGD_VALIDITY_UNKNOWN )
	{
		m_nDataHandlerCounter++;
		if ( m_nDataHandlerCounter >= BUFSIZ)
		{
			CPLError(CE_Failure, CPLE_AppDefined, "File probably corrupted (million laugh pattern)");
			XML_StopParser( m_Parser, XML_FALSE);
		}
	}

	if ( (m_nFlags & TAG_ATTRIBUTE) ||
		 ( (m_nFlags & TAG_DATE) && (m_nFlags & TAG_DATE_ATTRIBUTE) ) ||
		 (m_nFlags & TAG_COORDINATE) )
	{
		char szTemp[MAX_LINES];
		CPLStrlcpy( szTemp, data, MAX_LINES );
		szTemp[nLen] = '\0';
		CPLStrlcat( m_szTempChar, szTemp, MAX_LINES );
		m_nTempChar += nLen;
	}
}


bool FGDReader::parseNext()
{
	int nDone, nLen;
	char szBuf[BUFSIZ];
	XML_ParsingStatus status;

	if ( !m_Parser ) return false;

	XML_GetParsingStatus( m_Parser, &status );
	if ( status.parsing == XML_FINISHED )
		return false;

	{
		nLen = VSIFReadL( szBuf, 1, BUFSIZ, fp );
		nDone = VSIFEofL( fp );
		if ( XML_Parse( m_Parser, szBuf, nLen, nDone ) == XML_STATUS_ERROR )
		{
			fprintf(stderr, "Parse error at line %u:\n%s\n",
				XML_GetCurrentLineNumber( m_Parser ),
				XML_ErrorString(XML_GetErrorCode( m_Parser ) ) );
			return false;
		}
		if ( nDone )
			return false;
	}

	return true;

}


void FGDReader::ResetReading()
{
	m_dqpoFeature.clear();
	VSIFSeekL( fp, 0, SEEK_SET );
	m_nNextFID = 0;
	m_nFlags = 0x0000;

	m_poPoint = NULL;
	m_poLineString = NULL;
	m_poRing = NULL;
	m_poPolygon = NULL;

	if ( m_Parser )
		XML_ParserFree( m_Parser );

	m_Parser = XML_ParserCreate( NULL );
	XML_SetUserData( m_Parser, this );
	XML_SetElementHandler( m_Parser, ::startElement, ::endElement );
	XML_SetCharacterDataHandler( m_Parser, ::dataHandler );
	XML_SetUnknownEncodingHandler( m_Parser, XML_ShiftJISEncodingHandler, this );
}


OGRFeature* FGDReader::GetNextFeature()
{
	while ( !m_dqpoFeature.size() )
	{
		if ( !parseNext() )
		{
			break;
		}
	}

	if ( !m_dqpoFeature.size() )
		return NULL;

	OGRFeature *pFeature = m_dqpoFeature.front();
	m_dqpoFeature.pop_front();
	return pFeature;
}

#endif // #if defined(HAVE_EXPAT) && defined(HAVE_ICONV)
